# -*- mode: python; encoding: utf-8 -*-
#
# Copyright 2017 the Critic contributors, Opera Software ASA
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License.  You may obtain a copy of
# the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations under
# the License.

import logging
import passlib.context
import time

logger = logging.getLogger(__name__)

import api

name = "calibrate-pwhash"
description = "Calibrate password hashing."

def setup(critic, parser):
    parser.add_argument(
        "--hash-time", type=float, default=0.25,
        help="The target amount of time in seconds that hashing should take.")
    parser.set_defaults(need_session=True)

def main(critic, arguments):
    authentication_mode = api.critic.settings().frontend.authentication_mode
    if authentication_mode != "critic":
        logger.error(
            "Not meaningful with frontend.authentication_mode==\"%s\"",
            authentication_mode)
        return 1

    authdb = api.critic.settings().authentication.database
    if authdb != "internal":
        logger.error(
            "Not meaningful with authentication.database==\"%s\"", authdb)
        return 1

    hash_scheme = api.critic.settings().authentication.default_scheme

    min_rounds_name = "%s__min_rounds" % hash_scheme
    min_rounds_value = 100

    while True:
        calibration_context = passlib.context.CryptContext(
            schemes=[hash_scheme], default=hash_scheme,
            **{ min_rounds_name: min_rounds_value })

        before = time.process_time()

        for n in range(10):
            calibration_context.encrypt("password")

        # It's possible encryption was fast enough to measure as zero, or some
        # other ridiculously small number.  "Round" it up to at least one
        # millisecond for sanity.
        hash_time = max(0.001, (time.process_time() - before) / 10)

        if hash_time >= arguments.hash_time:
            break

        # Multiplication factor.  Make it at least 1.2, to ensure we actually
        # ever finish this loop, and at most 10, to ensure we don't over-shoot
        # by too much.
        factor = max(1.2, min(10.0, arguments.hash_time / hash_time))

        min_rounds_value = int(factor * min_rounds_value)

    logger.info("Hashing with min_rounds=%d took %.2f seconds",
                min_rounds_value, hash_time)

    setting = api.systemsetting.fetch(critic, "authentication.minimum_rounds")

    with api.transaction.Transaction(critic) as transaction:
        transaction.setSystemSetting(setting, min_rounds_value)
